#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
curvesmodel Model and view for new CurveItem configuration
"""
__all__ = ['TaurusCurveItemTableModel', 'CurveItemConf', 'CurveItemConfDlg']
#raise UnimplementedError('Under Construction!')

import copy

from taurus.external.qt import Qt, Qwt5
from guiqwt.styles import CurveParam, AxesParam, update_style_attr
from guiqwt.builder import make

import taurus
from taurus.core.taurusexception import TaurusException
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_ATTR_MIME_TYPE
from taurus.qt.qtgui.util.ui import UILoadable
from taurus.qt.qtgui.extra_guiqwt.styles import TaurusCurveParam

AXIS_ID2NAME = {Qwt5.QwtPlot.yLeft: 'left', Qwt5.QwtPlot.yRight: 'right',
                Qwt5.QwtPlot.xBottom: 'bottom', Qwt5.QwtPlot.xTop: 'top'}

# set some named constants
# columns:
NUMCOLS = 3
X, Y, TITLE = range(NUMCOLS)
SRC_ROLE = Qt.Qt.UserRole + 1


class Component(object):

    def __init__(self, src):
        self.display = ''
        self.icon = Qt.QIcon()
        self.ok = True
        self.processSrc(src)

    def processSrc(self, src):
        '''processes the src and sets the values of display, icon and ok attributes'''
        if src is None:
            self.display, self.icon, self.ok = '(Use indices)', Qt.QIcon(
            ), True
            return
        src = str(src).strip()
        # empty
        if src == '':
            self.display, self.icon, self.ok = '(Use indices)', Qt.QIcon(
            ), True
            return
        # for formulas
        if src.startswith('='):
            #@todo: evaluate/validate the expression
            self.display, self.icon, self.ok = src[
                1:].strip(), Qt.QIcon.fromTheme('accessories-calculator'), True
            return
        # for tango attributes
        try:
            attr = taurus.Attribute(src)
            dev = attr.getParentObj()
        except TaurusException:
            self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                'dialog-warning'), False
            return
        if not dev.isValidDev():
            self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                'network-error'), False
            return
        self.display, self.icon, self.ok = attr.getSimpleName(),\
                                           Qt.QIcon('logos:taurus.png'), True


class CurveItemConf(object):

    def __init__(self, taurusparam=None, curveparam=None, axesparam=None):
        if taurusparam is None:
            taurusparam = TaurusCurveParam()
        self.taurusparam = taurusparam
        if curveparam is None:
            curveparam = CurveParam()
            style = make.style.next()  # cycle through colors and linestyles
            update_style_attr(style, curveparam)
            curveparam.line.width = 2
        self.curveparam = curveparam
        self.axesparam = axesparam
        self.x = Component(taurusparam.xModel)
        self.y = Component(taurusparam.yModel)
        if not self.curveparam.label:
            self.curveparam.label = taurusparam.xModel

    def __repr__(self):
        ret = "CurveItemConf(xModel='%s', yModel='%s')" % (
            self.taurusparam.xModel, self.taurusparam.yModel)
        return ret

    @staticmethod
    def fromTaurusCurveItem(item):
        plot = item.plot()
        if plot is not None:
            axesparam = AxesParam()
            axesparam.update_param(item)
        return CurveItemConf(taurusparam=item.taurusparam, curveparam=item.curveparam, axesparam=axesparam)

    @staticmethod
    def fromAny(obj):
        '''return a CurveItemConf from whatever given in input (if possible).
        Raises ValueError if not possible'''
        if isinstance(obj, CurveItemConf):
            return copy.deepcopy(obj)
        try:
            return CurveItemConf.fromTaurusCurveItem(obj)
        except:
            raise
        try:
            return CurveItemConf(*obj)
        except:
            pass
        raise ValueError('Cannot convert %s into a CurveItemConf' % repr(obj))


class TaurusCurveItemTableModel(Qt.QAbstractTableModel):
    ''' A Qt data model for describing curves
    '''

    dataChanged = Qt.pyqtSignal('QModelIndex', 'QModelIndex')

    def __init__(self, curves=None):
        if curves is None:
            curves = []
        curves = [CurveItemConf.fromAny(c) for c in curves]  # convert curves
        super(TaurusCurveItemTableModel, self).__init__()
        self.ncolumns = NUMCOLS
        self.curves = curves

    def dumpData(self):
        return copy.deepcopy(self.curves)

    def rowCount(self, index=Qt.QModelIndex()):
        return len(self.curves)

    def columnCount(self, index=Qt.QModelIndex()):
        return self.ncolumns

    def data(self, index, role=Qt.Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return Qt.QVariant()
        row = index.row()
        column = index.column()
        # Display Role
        if role == Qt.Qt.DisplayRole:
            if column == X:
                return Qt.QVariant(str(self.curves[row].x.display))
            elif column == Y:
                return Qt.QVariant(str(self.curves[row].y.display))
            elif column == TITLE:
                return Qt.QVariant(str(self.curves[row].curveparam.label))
            else:
                return Qt.QVariant()
        elif role == Qt.Qt.DecorationRole:
            if column == X:
                return Qt.QVariant(self.curves[row].x.icon)
            elif column == Y:
                return Qt.QVariant(self.curves[row].y.icon)
            else:
                return Qt.QVariant()
        elif role == Qt.Qt.TextColorRole:
            if column == X:
                Qt.QVariant(
                    Qt.QColor(self.curves[row].x.ok and 'green' or 'red'))
            elif column == Y:
                Qt.QVariant(
                    Qt.QColor(self.curves[row].y.ok and 'green' or 'red'))
            else:
                return Qt.QVariant()
        elif role == SRC_ROLE:
            if column == X:
                return Qt.QVariant(str(self.curves[row].taurusparam.xModel))
            elif column == Y:
                return Qt.QVariant(str(self.curves[row].taurusparam.yModel))
            else:
                return Qt.QVariant()
        elif role == Qt.Qt.ToolTipRole:
            if column == X:
                return Qt.QVariant(str(self.curves[row].taurusparam.xModel))
            elif column == Y:
                return Qt.QVariant(str(self.curves[row].taurusparam.yModel))
            else:
                return Qt.QVariant()
        if role == Qt.Qt.EditRole:
            if column == X:
                return Qt.QVariant(str(self.curves[row].taurusparam.xModel))
            elif column == Y:
                return Qt.QVariant(str(self.curves[row].taurusparam.yModel))
            elif column == TITLE:
                return Qt.QVariant(str(self.curves[row].curveparam.label))
            else:
                return Qt.QVariant()
        return Qt.QVariant()

    def headerData(self, section, orientation, role=Qt.Qt.DisplayRole):
        if role == Qt.Qt.TextAlignmentRole:
            if orientation == Qt.Qt.Horizontal:
                return Qt.QVariant(int(Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter))
            return Qt.QVariant(int(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter))
        if role != Qt.Qt.DisplayRole:
            return Qt.QVariant()
        # So this is DisplayRole...
        if orientation == Qt.Qt.Horizontal:
            if section == X:
                return Qt.QVariant("X source")
            elif section == Y:
                return Qt.QVariant("Y Source")
            elif section == TITLE:
                return Qt.QVariant("Title")
            return Qt.QVariant()
        else:
            return Qt.QVariant(str(section + 1))

    def flags(self, index):  # use this to set the editable flag when fix is selected
        if not index.isValid():
            return Qt.Qt.ItemIsEnabled
        column = index.column()
        if column in (X, Y):
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled | Qt.Qt.ItemIsDropEnabled | Qt.Qt.ItemIsSelectable)
        elif column == TITLE:
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled)
        return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled)

    def setData(self, index, value=None, role=Qt.Qt.EditRole):
        if index.isValid() and (0 <= index.row() < self.rowCount()):
            row = index.row()
            curve = self.curves[row]
            column = index.column()
            value = Qt.from_qvariant(value, str)
            if column == X:
                curve.taurusparam.xModel = value
                curve.x.processSrc(value)
            elif column == Y:
                curve.taurusparam.yModel = value
                curve.y.processSrc(value)
            elif column == TITLE:
                curve.curveparam.label = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def insertRows(self, position=None, rows=1, parentindex=None):
        if position is None:
            position = self.rowCount()
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginInsertRows(parentindex, position, position + rows - 1)
        slice = [CurveItemConf() for i in range(rows)]
        self.curves = self.curves[:position] + slice + self.curves[position:]
        self.endInsertRows()
        return True

    def removeRows(self, position, rows=1, parentindex=None):
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginRemoveRows(parentindex, position, position + rows - 1)
        self.curves = self.curves[:position] + self.curves[position + rows:]
        self.endRemoveRows()
        self.reset()
        return True

    def clearAll(self):
        self.removeRows(0, self.rowCount())

    def mimeTypes(self):
        result = list(Qt.QAbstractTableModel.mimeTypes(self))
        result += [TAURUS_ATTR_MIME_TYPE, 'text/plain']
        return result

    def dropMimeData(self, data, action, row, column, parent):
        if row == -1:
            if parent.isValid():
                row = parent.row()
            else:
                row = parent.rowCount()
        if column == -1:
            if parent.isValid():
                column = parent.column()
            else:
                column = parent.columnCount()
        if data.hasFormat(TAURUS_ATTR_MIME_TYPE):
            self.setData(self.index(row, column),
                         value=Qt.QVariant(str(data.data(TAURUS_ATTR_MIME_TYPE))))
            return True
        elif data.hasFormat(TAURUS_MODEL_LIST_MIME_TYPE):
            models = str(data.data(TAURUS_MODEL_LIST_MIME_TYPE)).split()
            if len(models) == 1:
                self.setData(self.index(row, column),
                             value=Qt.QVariant(models[0]))
                return True
            else:
                self.insertRows(row, len(models))
                for i, m in enumerate(models):
                    self.setData(self.index(row + i, column),
                                 value=Qt.QVariant(m))
                return True
        elif data.hasText():
            self.setData(self.index(row, column), Qt.QVariant(data.text()))
            return True
        return False

    def mimeData(self, indexes):
        mimedata = Qt.QAbstractTableModel.mimeData(self, indexes)
        if len(indexes) == 1:
            #            data = Qt.from_qvariant(self.data(indexes[0], str)
            #            mimedata.setData(TAURUS_ATTR_MIME_TYPE, data)
            data = Qt.from_qvariant(self.data(indexes[0], role=SRC_ROLE), str)
            mimedata.setText(data)
        return mimedata
        # mimedata.setData()


@UILoadable(with_ui='ui')
class CurveItemConfDlg(Qt.QWidget):
    ''' A configuration dialog for creating new CurveItems.

    Provides a browser for Taurus models and an editable table for the sources
    and title of data
    '''

    dataChanged = Qt.pyqtSignal('QModelIndex', 'QModelIndex')
    applied = Qt.pyqtSignal()

    def __init__(self, parent=None, curves=None, showXcol=True):
        super(CurveItemConfDlg, self).__init__(parent)
        self.loadUi()
        self._showXcol = showXcol

        if curves is None:
            curves = [CurveItemConf()]
        # add the NeXusWidget if extra_nexus is available
        try:
            from taurus.qt.qtgui.extra_nexus import TaurusNeXusBrowser
            nexusWidget = TaurusNeXusBrowser()
            self.ui.tabWidget.addTab(nexusWidget, 'NeXus')
        except:
            import taurus.core.util.log
            _logger = taurus.core.util.log.Logger('CurveItemConfDlg')
            _logger.warning('TaurusNeXusBrowser not available')
            self.traceback()

        self.ui.tangoTree.setButtonsPos(Qt.Qt.RightToolBarArea)

        # @todo: The action for this button is not yet implemented
        self.ui.reloadBT.setEnabled(False)

        self.model = TaurusCurveItemTableModel(curves)

        table = self.ui.curvesTable
        table.setModel(self.model)
        table.setColumnHidden(X, not self._showXcol)

#        #adjust the column widths
#        table.resizeColumnsToContents()
#        availableSpace = table.viewport().width()
#        for col in xrange(self.model.columnCount()):
#            availableSpace -= table.columnWidth(col)
#        if availableSpace > 0:
#            extraSpace = availableSpace / self.model.columnCount()
#            for col in xrange(self.model.columnCount()):
#                table.setColumnWidth(col, table.columnWidth(col)+extraSpace)

        # host

        import taurus  # @todo: I get "UnboundLocalError: local variable 'taurus' referenced before assignment" if I don't import taurus again here ????
        host = taurus.Authority().getNormalName()
        self.ui.tangoTree.setModel(host)

        # Connections
        self.ui.applyBT.clicked.connect(self.onApply)
        self.ui.reloadBT.clicked.connect(self.onReload)
        self.ui.cancelBT.clicked.connect(self.close)
        self.ui.tangoTree.addModels.connect(self.onModelsAdded)
        self.ui.curvesTable.customContextMenuRequested.connect(self.onTableContextMenu)

    def onTableContextMenu(self, pos):
        index = self.ui.curvesTable.indexAt(pos)
        row = index.row()
        menu = Qt.QMenu(self.ui.curvesTable)
        if row >= 0:
            removeThisAction = menu.addAction(Qt.QIcon.fromTheme(
                'list-remove'), 'Remove this curve', self._onRemoveThisAction)
        removeAllAction = menu.addAction(Qt.QIcon.fromTheme(
            'edit-clear'), 'Clear all', self.model.clearAll)
        addRowAction = menu.addAction(Qt.QIcon.fromTheme(
            'list-add'), 'Add new row', self.model.insertRows)
        if row >= 0:
            editParsAction = menu.addAction(Qt.QIcon.fromTheme(
                'preferences-system'), 'Edit parameters of this curve...', self._onEditParsAction)

        menu.exec_(Qt.QCursor.pos())

    def _onRemoveThisAction(self):
        row = self.ui.curvesTable.currentIndex().row()
        self.model.removeRows(row)

    def _onEditParsAction(self):
        row = self.ui.curvesTable.currentIndex().row()
        if row >= 0:
            c = self.model.curves[row]
            params = [p for p in (c.curveparam, c.axesparam,
                                  c.taurusparam) if p is not None]
            if params:
                from guidata.dataset.datatypes import DataSetGroup
                group = DataSetGroup(
                    params, "Parameters for curve %s" % c.curveparam.label)
                group.edit()
                c.x.processSrc(c.taurusparam.xModel)
                c.y.processSrc(c.taurusparam.yModel)
                self.dataChanged.emit(self.model.index(
                    row, 0), self.model.index(row, self.model.rowCount()))

    def onModelsAdded(self, models):
        nmodels = len(models)
        rowcount = self.model.rowCount()
        self.model.insertRows(rowcount, nmodels)
        for i, m in enumerate(models):
            self.model.setData(self.model.index(
                rowcount + i, Y), value=Qt.QVariant(m))
            title = self.model.data(self.model.index(
                rowcount + i, Y))  # the display data
            self.model.setData(self.model.index(
                rowcount + i, TITLE), value=title)

    def getCurveItemConfs(self):
        return self.model.dumpData()

    @staticmethod
    def showDlg(parent=None, curves=None):
        '''Static method that launches a modal dialog containing a CurveItemConfDlg

        :param parent: (QObject) parent for the dialog

        :return: (list,bool or QMimeData,bool) Returns a models,ok tuple. models can be
                 either a list of models or a QMimeData object, depending on
                 `asMimeData`. ok is True if the dialog was accepted (by
                 clicking on the "update" button) and False otherwise
        '''
        dlg = Qt.QDialog(parent)
        dlg.setWindowTitle('Curves Selection')
        layout = Qt.QVBoxLayout()
        w = CurveItemConfDlg(parent=parent, curves=curves)
        layout.addWidget(w)
        dlg.setLayout(layout)
        w.applied.connect(dlg.accept)
        w.ui.cancelBT.clicked.connect(dlg.close)
        dlg.exec_()
        return w.getCurveItemConfs(), (dlg.result() == dlg.Accepted)

    def onApply(self):
        self.applied.emit()

    def onReload(self):
        print "RELOAD!!! (todo)"


#
#
# def test2():
#    from taurus.qt.qtgui.application import TaurusApplication
#    app = TaurusApplication(app_name="CurvesSelection", app_version=taurus.Release.version)
#
#    print CurveItemConfDlg.showDlg(curves= curves)
#
#    sys.exit()
#
#
# if __name__ == "__main__":
#    import sys
#    test2()



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

'''
evaluation module. See __init__.py for more detailed documentation
'''
__all__ = ['EvaluationFactory']


import weakref

from taurus.core.taurusbasetypes import TaurusElementType
from evalattribute import EvaluationAttribute
from evalauthority import EvaluationAuthority
from evaldevice import EvaluationDevice
from taurus.core.taurusexception import TaurusException
from taurus.core.tauruspollingtimer import TaurusPollingTimer
from taurus.core.util.log import Logger
from taurus.core.util.singleton import Singleton
from taurus.core.taurusfactory import TaurusFactory


class EvaluationFactory(Singleton, TaurusFactory, Logger):
    """
    A Singleton class that provides Evaluation related objects.
    """
    elementTypesMap = {TaurusElementType.Authority: EvaluationAuthority,
                       TaurusElementType.Device: EvaluationDevice,
                       TaurusElementType.Attribute: EvaluationAttribute
                       }
    schemes = ("eval", "evaluation")
    DEFAULT_DEVICE = '@DefaultEvaluator'
    DEFAULT_AUTHORITY = '//localhost'
    DEFAULT_DATABASE = '_DefaultEvalDB'

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args, **kwargs):
        """Singleton instance initialization."""
        name = self.__class__.__name__
        self.call__init__(Logger, name)
        self.call__init__(TaurusFactory)
        self.eval_attrs = weakref.WeakValueDictionary()
        self.eval_devs = weakref.WeakValueDictionary()
        self.eval_configs = weakref.WeakValueDictionary()
        self.scheme = 'eval'

    def findObjectClass(self, absolute_name):
        """Operation models are always OperationAttributes
        """
        if EvaluationDevice.isValid(absolute_name):
            return EvaluationDevice
        elif EvaluationAttribute.isValid(absolute_name):
            return EvaluationAttribute
        else:
            self.debug("Not able to find Object class for %s" % absolute_name)
            self.traceback()
            return None

    def getAuthority(self, name=None):
        """Obtain the EvaluationDatabase object.

        :param db_name: (str) this is ignored because only one database is supported

        :return: (EvaluationDatabase)
        """
        if name is None:
            name = 'eval://localhost'

        v = self.getAuthorityNameValidator()
        if not v.isValid(name):
            raise TaurusException(
                "Invalid Evaluation authority name %s" % name)

        if not hasattr(self, "_auth"):
            self._auth = EvaluationAuthority(self.DEFAULT_AUTHORITY)
        return self._auth

    def getDevice(self, dev_name):
        """Obtain the object corresponding to the given device name. If the
        corresponding device already exists, the existing instance is returned.
        Otherwise a new instance is stored and returned.

        :param dev_name: (str) the device name string. See
                         :mod:`taurus.core.evaluation` for valid device names

        :return: (EvaluationDevice)

        @throws TaurusException if the given name is invalid.
        """
        # try to see if the given name is already a known full_name
        d = self.eval_devs.get(dev_name, None)
        if d is None:
            # find the full_name and see if we already know it
            validator = self.getDeviceNameValidator()
            names = validator.getNames(dev_name)
            if names is None:
                raise TaurusException("Invalid evaluator device name %s" %
                                      dev_name)
            fullname, normalname, devname = names
            d = self.eval_devs.get(fullname, None)
            # if we do not know it, create the dev and store it in cache
            if d is None:
                groups = validator.getUriGroups(dev_name)
                if groups['_evalclass'] is not None:
                    modulename, classname = groups['_evalclass'].rsplit('.', 1)
                    try:
                        m = __import__(modulename, globals(), locals(),
                                       [classname], -1)
                        DevClass = getattr(m, classname)
                    except:
                        self.warning('Problem importing "%s"' % devname)
                        raise
                else:
                    DevClass = EvaluationDevice
                # Get authority (creating if necessary)
                auth_name = groups.get('authority') or self.DEFAULT_AUTHORITY
                authority = self.getAuthority(auth_name)
                # Create Device (and store it in cache via self._storeDev)
                d = DevClass(fullname, parent=authority,
                             storeCallback=self._storeDev)
        return d

    def getAttribute(self, attr_name, **kwargs):
        """Obtain the object corresponding to the given attribute name. If the
        corresponding attribute already exists, the existing instance is
        returned. Otherwise a new instance is stored and returned. The evaluator
        device associated to this attribute will also be created if necessary.

        :param attr_name: (str) the attribute name string. See
                          :mod:`taurus.core.evaluation` for valid attribute
                          names

        Any aditional keyword arguments will be passed directly to the
        constructor of `:class:EvaluationAttribute`

        :return: (EvaluationAttribute)

        @throws TaurusException if the given name is invalid.
        """
        a = self.eval_attrs.get(
            attr_name, None)  # first try with the given name
        if a is None:  # if not, try with the full name
            validator = self.getAttributeNameValidator()
            names = validator.getNames(attr_name)
            if names is None or names[0] is None:
                raise TaurusException(
                    "Invalid evaluation attribute name %s" % attr_name)
            fullname = names[0]
            a = self.eval_attrs.get(fullname, None)
            if a is None:  # if the full name is not there, create one
                dev = self.getDevice(validator.getDeviceName(attr_name))
                kwargs['storeCallback'] = self._storeAttr
                if not kwargs.has_key('pollingPeriod'):
                    kwargs['pollingPeriod'] = self.getDefaultPollingPeriod()
                a = EvaluationAttribute(fullname, parent=dev, **kwargs)
        return a

    def _storeDev(self, dev):
        name = dev.getFullName()
        exists = self.eval_devs.get(name)
        if exists is not None:
            if exists == dev:
                self.debug("%s has already been registered before" % name)
                raise DoubleRegistration
            else:
                self.debug(
                    "%s has already been registered before with a different object!" % name)
                raise DoubleRegistration
        self.eval_devs[name] = dev

    def _storeAttr(self, attr):
        name = attr.getFullName()
        exists = self.eval_attrs.get(name)
        if exists is not None:
            if exists == attr:
                self.debug("%s has already been registered before" % name)
                raise DoubleRegistration
            else:
                self.debug(
                    "%s has already been registered before with a different object!" % name)
                raise DoubleRegistration
        self.eval_attrs[name] = attr

    def _storeConfig(self, fullname, config):
        #name = config.getFullName()
        name = fullname
        exists = self.eval_configs.get(name)
        if exists is not None:
            if exists == config:
                self.debug("%s has already been registered before" % name)
                raise DoubleRegistration
            else:
                self.debug(
                    "%s has already been registered before with a different object!" % name)
                raise DoubleRegistration
        self.eval_configs[name] = config

    def addAttributeToPolling(self, attribute, period, unsubscribe_evts=False):
        """Activates the polling (client side) for the given attribute with the
           given period (seconds).

           :param attribute: (taurus.core.tango.TangoAttribute) attribute name.
           :param period: (float) polling period (in seconds)
           :param unsubscribe_evts: (bool) whether or not to unsubscribe from events
        """
        tmr = self.polling_timers.get(period, TaurusPollingTimer(period))
        self.polling_timers[period] = tmr
        tmr.addAttribute(attribute, self.isPollingEnabled())

    def removeAttributeFromPolling(self, attribute):
        """Deactivate the polling (client side) for the given attribute. If the
           polling of the attribute was not previously enabled, nothing happens.

           :param attribute: (str) attribute name.
        """
        p = None
        for period, timer in self.polling_timers.iteritems():
            if timer.containsAttribute(attribute):
                timer.removeAttribute(attribute)
                if timer.getAttributeCount() == 0:
                    p = period
                break
        if p:
            del self.polling_timers[period]

    def getAuthorityNameValidator(self):
        """Return EvaluationAuthorityNameValidator"""
        import evalvalidator
        return evalvalidator.EvaluationAuthorityNameValidator()

    def getDeviceNameValidator(self):
        """Return EvaluationDeviceNameValidator"""
        import evalvalidator
        return evalvalidator.EvaluationDeviceNameValidator()

    def getAttributeNameValidator(self):
        """Return EvaluationAttributeNameValidator"""
        import evalvalidator
        return evalvalidator.EvaluationAttributeNameValidator()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
.. currentmodule:: taurus.core.resource

Resource scheme extension for taurus core mode.
The resource extension is a special extension that acts like a name map
for actual model names.

This allows to indirect the hardcoded model names of your application and keep
the actual specific model names grouped in one place for better portability and
reusability of your application.

The scheme name is 'res'. The map can be implemented either as python
modules or as dicts (see below).

The main class for the extension module is :class:`ResourcesFactory` and
you can add resource maps with :meth:`ResourcesFactory.loadResource`.

By default, the :class:`ResourcesFactory` will always try to load a resource
module named 'taurus_resources' from the application directory (so
you can create a file called `taurus_resources.py` in your application directory
and skip the step of calling :meth:`ResourcesFactory.loadResource`.


Mapping implemented as python modules
-------------------------------------

If a resource is a python module, the factory will use its global variables as
the resource keys. The variable value, which must be a string will be used as
the mapped model. For example, if the `taurus_resources.py` file in
the application directory contains the following definitions::

    my_device = 'tango:a/b/c'
    my_state = my_device + '/state'

Then, in your code, you can access the Device and Attribute objects by doing::

    >>> import taurus
    >>> my_device_obj = taurus.Device('res:my_device')
    >>> my_state_obj = taurus.Attribute('res:my_state')

Note that you can use python code to automate the contents of the module.
Example::

    base = 'my/motor/'
    g = globals()

    for i in xrange(256):
        i_str = str(i)
        g['mym"+i_str] = base + i_str

Mapping implemented as dictionaries
-----------------------------------

Dictionaries can also be registered (as an alternative to modules) as resource
maps::

    >>> d = {'my_device':'tango:a/b/c', 'my_state':'tango:a/b/c/state'}
    >>> import taurus
    >>> factory = taurus.Factory('res')
    >>> factory.loadResource(d)
    >>> my_device_obj = taurus.Device('res:my_device')
    >>> my_state_obj = taurus.Attribute('res:my_state')

Note: for coherence with the file mapping, valid key names are restricted to
valid variable names (i.e. to the following regexp: `[a-zA-Z_][a-zA-Z0-9_]*`

Important: Models object type is the mapped type
-------------------------------------------------

The model objects returned by the :class:`ResourcesFactory` will be of the
mapped model type. In other words: if the model name 'eval:Q(rand())' is
mapped to the `myattr` resource name, calling `taurus.Attribute('res:myattr')`
will return a :class:`EvaluationAttribute`, not a `ResAttribute` (`ResAttribute`
is not even defined).

"""

from resfactory import *



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ['EvaluationDeviceNameValidator',
           'EvaluationAttributeNameValidator']

import re
import hashlib

import taurus
from taurus import isValidName, debug
from taurus.core import TaurusElementType

from taurus.core.taurusvalidator import (TaurusAttributeNameValidator,
                                         TaurusDeviceNameValidator,
                                         TaurusAuthorityNameValidator)

# Pattern for python variables
PY_VAR = r'(?<![\.a-zA-Z0-9_])[a-zA-Z_][a-zA-Z0-9_]*'
PY_VAR_RE = re.compile(PY_VAR)
# Pattern for semicolon-separated <variable>=<value> pairs (in URI paths)
K_EQUALS_V = r'(%s)=([^?#=;]+)' % PY_VAR
K_EQUALS_V_RE = re.compile(K_EQUALS_V)
#
QUOTED_TEXT = '(".*?"|\'.*?\')'
QUOTED_TEXT_RE = re.compile(QUOTED_TEXT)


def _findAllTokensBetweenChars(string, start, end, n=None):
    '''Finds the text between (possibly nested) delimiters in a string.
    In case of nested delimiters, only the outermost level is
    returned. It returns a tuple of (idx,token)

    Example::

      _findAllTokensBetweenChars('{foo}bar{zig{zag}}boom', '{', '}')
      --> [(1,'foo'), (9, 'zig{zag}')]


    :param string: (str) the expression to parse
    :param start: (str) the char delimiting the start of a token
    :param end: (str) the char delimiting the end of a token
    :param n: (int or None) If an int is passed, it sets the maximum number
              of tokens to be found

    :return: (list(<int>,<str>)) a list of (idx, token) tuples. The idx is the
    position of the token in `string`
             (tokens d not include the delimiting chars not including the
             brackets)
    '''
    if start == end:
        raise ValueError('star_char must be different from end_char')
    if string.count(start) != string.count(end):
        raise ValueError('Non-matching delimiters (%i "%s" vs %i "%s")' %
                         string.count(start), start, string.count(end), end)
    tokens = []
    indexes = []
    idx = 0
    rest = string
    while len(tokens) != n:
        s = rest.find(start)
        if s < 0:
            break
        e = rest.find(end) + 1
        while rest[s:e].count(start) != rest[s:e].count(end):
            ne = rest[e:].find(end)
            e = e + 1 + ne
        tokens.append((idx + s, rest[s + 1:e - 1]))
        idx += e
        rest = rest[e:]
    return tokens


def _isQuoted(string, substring, idx):
    """returns True if position i of string is in a quoted region"""
    bfr = string[:idx]
    aft = string[idx + len(substring):]
    if (bfr.count('"') % 2 or aft.count('"') % 2 or
            bfr.count("'") % 2 or aft.count("'") % 2):
        return True
    else:
        return False


def _replacepos(string, old, new, idx):
    """return copy of string where the occurrence of substring `old` at
    position `pos` is replaced by `new`
    """
    if not string[idx:].startswith(old):
        raise Exception('invalid')
    return string[:idx] + new + string[idx + len(old):]


class EvaluationAuthorityNameValidator(TaurusAuthorityNameValidator):
    '''Validator for Evaluation authority names. For now, the only supported
    authority (in strict mode) is "//localhost":
    '''
    scheme = 'eval'
    authority = '//localhost'
    path = '(?!)'
    query = '(?!)'
    fragment = '(?!)'

    @property
    def nonStrictNamePattern(self):
        '''implement in derived classes if a "less strict" pattern is allowed
        (e.g. for backwards-compatibility, "tango://a/b/c" could be an accepted
        device name, even if it breaks RFC3986).
        '''
        return r'^(?P<scheme>eval|evaluation)://(db=(?P<dbname>[^?#;]+))$'


class EvaluationDeviceNameValidator(TaurusDeviceNameValidator):
    '''Validator for Evaluation device names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - devname: device name (either _evalname or _evalclass)
     - [_evalname]: evaluation instance name (aka non-dotted dev name)
     - [_evalclass]: evaluator class name (if dotted name given)
     - [_old_devname]: devname without "@". Only in non-strict mode
     - [_dbname] and [_subst]: unused. Only if non-strict mode

    Note: brackets on the group name indicate that this group will only contain
    a string if the URI contains it.
    '''

    scheme = 'eval'
    authority = EvaluationAuthorityNameValidator.authority
    _evaluatorname = r'((?P<_evalname>[^/?#:\.=]+)|(?P<_evalclass>(\w+\.)+\w+))'
    devname = r'(?P<devname>@%s)' % _evaluatorname
    path = r'(?!//)/?%s' % devname
    query = '(?!)'
    fragment = '(?!)'

    def getUriGroups(self, name, strict=None):
        '''reimplemented from :class:`TaurusDeviceNameValidator` to provide
        backwards compatibility with ol syntax'''
        groups = TaurusDeviceNameValidator.getUriGroups(self, name,
                                                        strict=strict)
        if groups is not None and not groups['__STRICT__']:
            _old_devname = groups['_old_devname']
            groups['devname'] = '@%s' % _old_devname
            if '.' in _old_devname:
                groups['_evalname'] = None
                groups['_evalclass'] = _old_devname
            else:
                groups['_evalname'] = _old_devname
                groups['_evalclass'] = None
        return groups

    def getNames(self, fullname, factory=None):
        '''reimplemented from :class:`TaurusDeviceNameValidator`'''
        from evalfactory import EvaluationFactory
        # TODO: add mechanism to select strict mode instead of hardcoding here
        groups = self.getUriGroups(fullname)
        if groups is None:
            return None

        authority = groups.get('authority')
        if authority is None:
            f_or_fklass = factory or EvaluationFactory
            groups['authority'] = authority = f_or_fklass.DEFAULT_AUTHORITY

        complete = 'eval:%(authority)s/%(devname)s' % groups
        normal = '%(devname)s' % groups
        short = normal.lstrip('@')

        return complete, normal, short

    @property
    def nonStrictNamePattern(self):
        '''In non-strict mode support old-style eval names
        '''
        p = r'^(?P<scheme>eval|evaluation)://(db=(?P<_dbname>[^?#;]+);)?' + \
            r'(dev=(?P<_old_devname>%s))' % self._evaluatorname + \
            r'(\?(?!configuration=)(?P<_subst>[^#?]*))?$'
        return p


class EvaluationAttributeNameValidator(TaurusAttributeNameValidator):
    '''Validator for Evaluation attribute names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - attrname: attribute name. same as concatenating _subst with _expr
     - _expr: a mathematical expression
     - _evalrefs: a list of eval refs found in the name (see :meth:`getRefs`)
     - [_subst]: a semicolon-separated repetition of key=value (for replacing
       them in _expr)
     - [devname]: as in :class:`EvaluationDeviceNameValidator`
     - [_evalname]: evaluation instance name (aka non-dotted dev name)
     - [_evalclass]: evaluator class name (if dotted name given)
     - [_old_devname]: devname without "@". Only in non-strict mode
     - [_dbname] and [_subst]: unused. Only if non-strict mode
     - [cfgkey] same as fragment (for bck-compat use only)

    Note: brackets on the group name indicate that this group will only contain
    a value if the URI contains it.
    '''
    scheme = 'eval'
    authority = EvaluationAuthorityNameValidator.authority
    path = ((r'(?!//)/?(%s/)?' +
             r'(?P<attrname>(?P<_subst>(%s;)+)?(?P<_expr>[^@?#]+))') %
            (EvaluationDeviceNameValidator.devname, K_EQUALS_V)
            )
    query = '(?!)'
    fragment = '(?P<cfgkey>[^# ]*)'

    @staticmethod
    def expandExpr(expr, substmap):
        '''expands expr by substituting all keys in map by their value.
        Note that eval references in expr (i.e. text within curly brackets)
        is not substituted.

        :param expr: (str) string that may contain symbols defined in symbolMap
        :param symbolMap: (dict or str) dictionary whose keys (strings) are
                          symbols to be substituted in `expr` and whose values
                          are the corresponding replacements. Alternatively, a
                          string containing a semi-colon separated list of
                          symbol=value pairs can also be passed.
        '''
        if isinstance(substmap, str):
            substmap = dict(K_EQUALS_V_RE.findall(substmap))
        ret = expr
        protected = {}

        # temporarily replace the text within quotes by hash-based placeholders
        for s in QUOTED_TEXT_RE.findall(expr):
            placeholder = hashlib.md5(s).hexdigest()
            protected[placeholder] = s
            ret = re.sub(s, placeholder, ret)

        # Substitute each k by its v in the expr (unless they are in
        # references)
        for k, v in substmap.iteritems():
            # create a pattern for matching complete word k
            # unless it is within between curly brackets
            keyPattern = r'(?<!\w)%s(?!\w)(?![^\{]*\})' % k
            # substitute matches of keyPattern by their value
            ret = re.sub(keyPattern, v, ret)

        # restore the protected strings
        for placeholder, s in protected.iteritems():
            ret = re.sub(placeholder, s, ret)
        return ret

    @staticmethod
    def getRefs(expr, ign_quoted=True):
        '''Find the attribute references (strings within brackets) in an eval
        expression. In case of nested references, only the outermost level is
        returned.

        Example: val.getRefs('{foo}bar{zig{zag}}boom') --> ['foo', 'zig{zag}']

        :param expr: (str) the expression to parse
        :param ign_quoted: If True (default) ignore refs within quotes

        :return (list<str>) a list of refs (not including the brackets)
        '''
        refs = _findAllTokensBetweenChars(expr, '{', '}')

        # print expr
        # print '0123456789012345678901234567890123456789'
        # print refs

        if refs and not ign_quoted:
            _, refs = zip(*refs)
            return refs

        ret = []
        for i, ref in refs:
            if not _isQuoted(expr, "{" + ref + "}", i):
                ret.append(ref)
        return ret

    @staticmethod
    def replaceUnquotedRef(string, substring, repl):
        """Return a copy of string where first non-quoted occurrence of
        `substring`
         is replaced by `repl`

        :param string: (str) string to be used
        :param substring: (str) substring to be replaced
        :param repl: (str) replacement

        :return: (str)

        """
        idx = string.find(substring)
        while _isQuoted(string, substring, idx):
            idx = string.find(substring, idx + 1)
        return _replacepos(string, substring, repl, idx)

    def isValid(self, name, matchLevel=None, strict=None):
        '''reimplemented from :class:`TaurusAttributeNameValidator` to do extra
        check on references validity (recursive)
        '''
        # Standard implementation
        if matchLevel is not None:
            groups = self._isValidAtLevel(name, matchLevel=matchLevel)
        else:
            groups = self.getUriGroups(name, strict=strict)
        if groups is None:
            return False

        # now check the references
        for ref in groups['_evalrefs']:
            if not isValidName(ref, etypes=(TaurusElementType.Attribute,),
                               strict=strict):
                debug('"%s" is invalid because ref "%s" is not a ' +
                      'valid attribute', name, ref)
                return False
        return True

    def getUriGroups(self, name, strict=None):
        '''reimplemented from :class:`TaurusAttributeNameValidator` to provide
        backwards compatibility with old syntax'''

        # mangle refs before matching the pattern to sanitize them
        refs = self.getRefs(name, ign_quoted=False)
        refs_dict = {}
        _name = name
        for i, ref in enumerate(refs):
            refs_dict['__EVALREF_%d__' % i] = '{%s}' % ref
            _name = _name.replace('{%s}' % ref, '{__EVALREF_%d__}' % i, 1)

        _groups = TaurusAttributeNameValidator.getUriGroups(self, _name,
                                                            strict=strict)
        if _groups is None:
            return None

        # create the groups dict with unmangled refs in its values
        groups = {}
        for n, g in _groups.items():
            if isinstance(g, str):  # avoid None or boolean values
                g = g.format(**refs_dict)
            groups[n] = g

        if not groups['__STRICT__']:
            # adapt attrname to what would be in strict mode
            _subst = groups['_subst'] or ''
            _expr = groups['_expr']
            if _subst:
                groups['attrname'] = "%s;%s" % (_subst.rstrip(';'), _expr)
            else:
                groups['attrname'] = _expr

            # adapt devname to what would be in strict mode
            old_devname = groups['_old_devname']
            if old_devname is None:
                groups['devname'] = None
            else:
                groups['devname'] = '@%s' % old_devname

        # check that there are not ";" in the expr (ign. quoted text and refs)
        sanitized_expr = QUOTED_TEXT_RE.sub('', groups['_expr'])
        for ref in self.getRefs(sanitized_expr, ign_quoted=False):
            sanitized_expr = sanitized_expr.replace(ref, '')
        if ";" in sanitized_expr:
            return None

        # add a group containing refs in attrname (ign. those in quoted text)
        groups['_evalrefs'] = self.getRefs(groups['attrname'], ign_quoted=True)

        return groups

    def _getSimpleNameFromExpression(self, expression):
        """Get the simple name of an evaluationAttribute from an expression"""
        name = expression
        for ref in self.getRefs(expression, ign_quoted=True):
            manager = taurus.core.TaurusManager()
            scheme = manager.getScheme(ref)
            _f = taurus.Factory(scheme)
            attrNameValidator = _f.getAttributeNameValidator()
            _, _, simple_name = attrNameValidator.getNames(ref)
            name = self.replaceUnquotedRef(name, '{%s}' % ref, simple_name)
        return name

    def _expandRefNames(self, attrname):
        """Expand the refs in an eval name to their full names"""
        name = attrname
        for ref in self.getRefs(attrname, ign_quoted=True):
            manager = taurus.core.TaurusManager()
            scheme = manager.getScheme(ref)
            _f = taurus.Factory(scheme)
            attrNameValidator = _f.getAttributeNameValidator()
            full_name, _, _ = attrNameValidator.getNames(ref)
            if full_name is None:
                debug('Cannot expand the fullname of %s' % ref)
                return None
            name = self.replaceUnquotedRef(name, '{%s}' % ref,
                                           '{%s}' % full_name)
        return name

    def getNames(self, fullname, factory=None, fragment=False):
        '''reimplemented from :class:`TaurusDeviceNameValidator`'''
        from evalfactory import EvaluationFactory
        groups = self.getUriGroups(fullname)
        if groups is None:
            return None

        f_or_fklass = factory or EvaluationFactory

        authority = groups.get('authority')
        if authority is None:
            groups['authority'] = authority = f_or_fklass.DEFAULT_AUTHORITY

        devname = groups.get('devname')
        if devname is None:
            groups['devname'] = devname = f_or_fklass.DEFAULT_DEVICE

        complete = 'eval:%s/%s/%s' % (authority, devname, groups['attrname'])
        complete = self._expandRefNames(complete)

        normal = groups['attrname']
        if devname != f_or_fklass.DEFAULT_DEVICE:
            normal = '%s/%s' % (devname, normal)
        if authority != f_or_fklass.DEFAULT_AUTHORITY:
            normal = '%s/%s' % (authority, normal)
        short = self._getSimpleNameFromExpression(groups['_expr'])

        # return fragment if requested
        if fragment:
            key = groups.get('fragment', None)
            return complete, normal, short, key
        return complete, normal, short

    @property
    def nonStrictNamePattern(self):
        '''In non-strict mode support old-style eval config names
        '''
        p = r'^(?P<scheme>eval|evaluation)://(db=(?P<_dbname>[^?#;]+);)?' + \
            r'(dev=(?P<_old_devname>[^?#;]+);)?' + \
            r'(?P<_expr>[^?#;]+)' + \
            r'(\?(?P<_substquery>(?!configuration=)(?P<_subst>%s(;%s)*)))?' % \
            (K_EQUALS_V, K_EQUALS_V) + \
            r'(\?(?P<query>configuration(=' + \
            '(?P<fragment>(?P<cfgkey>[^#?]*)))?))?$'

        return p

    def getExpandedExpr(self, name):
        '''
        Returns the expanded expression from the attribute name URI

        :param name: (str) eval attribute URI

        :return: (str) the expression (from the name )expanded with any
                 substitution k,v pairs also defined in the name
        '''
        groups = self.getUriGroups(name)
        if groups is None:
            return None
        _expr = groups['_expr']
        _subst = groups['_subst']
        return self.expandExpr(_expr, _subst or {})

    def getAttrName(self, s):
        #@TODO: Maybe this belongs to the factory, not the validator
        # TODO: this is pre-tep14 API from the
        # EvaluationConfigurationNameValidator. Check usage and remove.
        names = self.getNames(s)
        if names is None:
            return None
        return names[0]

    def getDeviceName(self, name):
        #@TODO: Maybe this belongs to the factory, not the validator
        '''Obtain the fullname of the device from the attribute name'''
        from evalfactory import EvaluationFactory
        groups = self.getUriGroups(name)
        if groups is None:
            return None
        authority = groups.get('authority')
        if authority is None:
            authority = EvaluationFactory.DEFAULT_AUTHORITY
        devname = groups.get('devname')
        if devname is None:
            devname = EvaluationFactory.DEFAULT_DEVICE
        return 'eval:%s/%s' % (authority, devname)

    def getDBName(self, s):
        #@TODO: Maybe this belongs to the factory, not the validator
        '''returns the full data base name for the given attribute name'''
        from evalfactory import EvaluationFactory
        m = self.name_re.match(s)
        if m is None:
            return None
        dbname = m.group('dbname') or EvaluationFactory.DEFAULT_DATABASE
        return "eval://db=%s" % dbname


if __name__ == '__main__':
    pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
qbuttonbox.py:
"""

__all__ = ["QButtonBox"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt


class QButtonBox(Qt.QDialogButtonBox):

    # PyQt Signals

    okClicked = Qt.pyqtSignal()
    openClicked = Qt.pyqtSignal()
    saveClicked = Qt.pyqtSignal()
    cancelClicked = Qt.pyqtSignal()
    closeClicked = Qt.pyqtSignal()
    discardClicked = Qt.pyqtSignal()
    applyClicked = Qt.pyqtSignal()
    resetClicked = Qt.pyqtSignal()
    restoreDefaultsClicked = Qt.pyqtSignal()
    helpClicked = Qt.pyqtSignal()
    resetClicked = Qt.pyqtSignal()
    saveAllClicked = Qt.pyqtSignal()
    yesClicked = Qt.pyqtSignal()
    yesToAllClicked = Qt.pyqtSignal()
    noClicked = Qt.pyqtSignal()
    abortClicked = Qt.pyqtSignal()
    retryClicked = Qt.pyqtSignal()
    ignoreClicked = Qt.pyqtSignal()

    def __init__(self, parent=None, designMode=False, buttons=None,
                 orientation=Qt.Qt.Horizontal):

        if buttons is None:
            buttons = Qt.QDialogButtonBox.Ok | Qt.QDialogButtonBox.Cancel

        Qt.QDialogButtonBox.__init__(self, buttons, orientation, parent)

        self.clicked.connect(self.onClicked)

    def onClicked(self, button):
        if self.standardButton(button) == Qt.QDialogButtonBox.Ok:
            self.okClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Open:
            self.openClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Save:
            self.saveClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Cancel:
            self.cancelClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Close:
            self.closeClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Discard:
            self.discardClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Apply:
            self.applyClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Reset:
            self.resetClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.RestoreDefaults:
            self.restoreDefaultsClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Help:
            self.helpClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.SaveAll:
            self.saveAllClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Yes:
            self.yesClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.YesToAll:
            self.yesToAllClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.No:
            self.noClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.NoToAll:
            self.noToAllClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Abort:
            self.abortClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Retry:
            self.retryClicked.emit()
        elif self.standardButton(button) == Qt.QDialogButtonBox.Ignore:
            self.ignoreClicked.emit()


if __name__ == "__main__":
    import sys

    def on_ok():
        print('OK!')

    def on_cancel():
        print('Cancel!')

    app = Qt.QApplication(sys.argv)
    bb = QButtonBox()
    bb.okClicked.connect(on_ok)
    bb.cancelClicked.connect(on_cancel)
    bb.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the set of base classes designed to provide
configuration features to the classes that inherit from them"""

__all__ = ["configurableProperty", "BaseConfigurableClass"]

__docformat__ = 'restructuredtext'


class configurableProperty:
    '''A dummy class used to handle properties with the configuration API

    .. warning:: this class is intended for internal use by the configuration
                 package. Do not instantiate it directly in your code.
                 Use :meth:`BaseConfigurableClass.registerConfigProperty` instead.
    '''

    def __init__(self, name, fget, fset, obj=None):
        self.name = name
        self.fget = fget  # this may either be a method or a method name
        self.fset = fset  # this may either be a method or a method name
        self._obj = obj  # obj is only needed if fset or fget are method names

    def createConfig(self, allowUnpickable=False):
        '''returns value returned by the fget function of this property. the allowUnpickable parameter is ignored'''
        if isinstance(self.fget, basestring):  # fget is not a method but a method name...
            result = getattr(self._obj, self.fget)()
        else:
            result = self.fget()
        return result

    def applyConfig(self, value, depth=-1):
        '''calls the fset function for this property with the given value. The depth parameter is ignored'''
        if isinstance(self.fget, basestring):  # fget is not a method but a method name...
            getattr(self._obj, self.fset)(value)
        else:
            self.fset(value)

    def objectName(self):
        '''returns the name of this property'''
        return self.name


class BaseConfigurableClass:
    '''
    A base class defining the API for configurable objects.

    .. note:: One implicit requisite is that a configurable object must also provide a
              `meth:`objectName` method which returns the object name. This is typically
              fulfilled by inheriting from QObject.

    Using objects that inherit from :class:`BaseConfigurableClass` automates
    saving and restoring of application settings and also enables the use of
    perspectives in Taurus GUIs.

    The basic idea is that each object/widget in your application is responsible
    for providing a dictionary containing information on its properties (see
    :meth:`createConfig`). The same object/widget is also responsible for
    restoring such properties when provided with a configuration dictionary (see
    :meth:`applyConfig`).

    For a certain property to be saved/restored it is usually enough to
    *register* it using :meth:`registerConfigProperty`. When the objects are
    structured in a hierarchical way (e.g. as the widgets in a Qt application),
    the parent widget can (should) delegate the save/restore of its children to
    the children themselves. This delegation is done by registering the children
    using :meth:`registerConfigDelegate`.

    Consider the following example: I am creating a groupbox container which
    contains a :class:`TaurusForm` and I want to save/restore the state of the
    checkbox and the properties of the form::

        #The class looks like this:
        class MyBox(Qt.QGroupBox, BaseConfigurableClass):
            def __init__(self):
                ...
                self.form = TaurusForm()
                ...
                self.registerConfigProperty(self.isChecked, self.setChecked, 'checked')
                self.registerConfigDelegate(self.form)   #the TaurusForm already handles its own configuration!
                ...

        #and we can retrieve the configuration doing:
        b1 = MyBox()
        b1.setChecked(True)  #checked is a registered property of MyBox class
        b1.form.setModifiableByUser(True)  #modifiableByUser is a registered property of a TaurusForm
        cfg = b1.createConfig()  #we get the configuration as a dictionary
        ...
        b2 = MyBox()
        b2.applyConfig(cfg)  #now b2 has the same configuration as b1 when cfg was created

    :meth:`createConfig` and :meth:`applyConfig` methods use a dictionary for
    passing the configuration, but :class:`BaseConfigurableClass` also provides
    some other convenience methods for working with files
    (:meth:`saveConfigFile` and :meth:`loadConfigFile`) or as QByteArrays
    (:meth:`createQConfig` and :meth:`applyQConfig`)

    Finally, we recommend to use :class:`TaurusMainWindow` for all Taurus GUIs
    since it automates all the steps for *saving properties when closing* and
    *restoring the settings on startup*. It also provides a mechanism for
    implementing "perspectives" in your application.

    '''

    defaultConfigRecursionDepth = -1
    # the latest element of this list is considered the current version
    _supportedConfigVersions = ("__UNVERSIONED__",)

    def __init__(self):
        self.resetConfigurableItems()

    @staticmethod
    def isTaurusConfig(x):
        '''Checks if the given argument has the structure of a configdict

        :param x: (object) object to test

        :return: (bool) True if it is a configdict, False otherwise.
        '''
        if not isinstance(x, dict):
            return False
        for k in ('__orderedConfigNames__', '__itemConfigurations__', 'ConfigVersion', '__pickable__'):
            if k not in x:
                return False

        for k in x['__orderedConfigNames__']:
            if k not in x['__itemConfigurations__']:
                print 'missing configuration for "%s" in %s' % (k, repr(x))
        return True

    def createConfig(self, allowUnpickable=False):
        '''
        Returns a dictionary containing configuration information about the
        current state of the object.

        In most usual situations, using :meth:`registerConfigProperty` and
        :meth:`registerConfigDelegate`, should be enough to cover all needs using
        this method, although it can be reimplemented in children classes to support
        very specific configurations.

        By default, meth:`createQConfig` and meth:`saveConfigFile` call to this
        method for obtaining the data.

        Hint: The following code allows you to serialize the configuration
        dictionary as a string (which you can store as a QSetting, or as a Tango
        Attribute, provided that allowUnpickable==False)::

            import pickle
            s = pickle.dumps(widget.createConfig())  #s is a string that can be stored

        :param alllowUnpickable:  (bool) if False the returned dict is
                                  guaranteed to be a pickable object. This is
                                  the default and preferred option because it
                                  allows the serialization as a string that can
                                  be directly stored in a QSetting. If True, this
                                  limitation is not enforced, which allows to
                                  use more complex objects as values (but limits
                                  its persistence).

        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).

        .. seealso: :meth:`applyConfig` , :meth:`registerConfigurableItem`,  meth:`createQConfig`, meth:`saveConfigFile`
        '''
        configdict = {"ConfigVersion": self._supportedConfigVersions[-1],
                      "__pickable__": True}
        # store the configurations for all registered configurable items as
        # well
        itemcfgs = {}
        for k, v in self.__configurableItems.iteritems():
            itemcfgs[k] = v.createConfig(allowUnpickable=allowUnpickable)
        configdict["__itemConfigurations__"] = itemcfgs
        configdict["__orderedConfigNames__"] = self.__configurableItemNames
        return configdict

    def applyConfig(self, configdict, depth=None):
        """applies the settings stored in a configdict to the current object.

        In most usual situations, using :meth:`registerConfigProperty` and
        :meth:`registerConfigDelegate`, should be enough to cover all needs using
        this method, although it can be reimplemented in children classes to support
        very specific configurations.

        :param configdict: (dict)
        :param depth: (int)  If depth = 0, applyConfig will only be called
                      for this object, and not for any other object registered
                      via :meth:`registerConfigurableItem`. If depth > 0,
                      applyConfig will be called recursively as many times as
                      the depth value. If depth < 0 (default, see note), no
                      limit is imposed to recursion (i.e., it will recurse for
                      as deep as there are registered items).

        .. note:: the default recursion depth can be tweaked in derived classes
                  by changing the class property `defaultConfigRecursionDepth`

        .. seealso:: :meth:`createConfig`
        """
        if depth is None:
            depth = self.defaultConfigRecursionDepth
        if not self.checkConfigVersion(configdict):
            raise ValueError(
                'the given configuration is of unsupported version')
        # delegate restoring the configuration of any registered configurable
        # item
        if depth != 0:
            itemcfgs = configdict["__itemConfigurations__"]
            # we use the sorted item names that was stored in the configdict
            for key in configdict["__orderedConfigNames__"]:
                if key in self.__configurableItems:
                    self.__configurableItems[key].applyConfig(
                        itemcfgs[key], depth=depth - 1)

    def getConfigurableItemNames(self):
        '''returns an ordered list of the names of currently registered
        configuration items (delegates and properties)

        :return: (list<unicode>)
        '''
        return self.__configurableItemNames

    def resetConfigurableItems(self):
        ''' clears the record of configurable items depending of this object

        .. seealso:: :meth:`registerConfigurableItem`
        '''
        self.__configurableItemNames = []
        self.__configurableItems = {}

    def registerConfigurableItem(self, item, name=None):
        print "Deprecation WARNING: %s.registerConfigurableItem() has been deprecated. Use registerConfigDelegate() instead" % repr(self)
        self._registerConfigurableItem(item, name=name)

    def registerConfigDelegate(self, delegate, name=None):
        '''
        Registers the given object as a delegate for configuration.
        Delegates are typically other objects inheriting from BaseConfigurableClass
        (or at least they must provide the following methods:

          - `createConfig` (as provided by, e.g., BaseConfigurableClass)
          - `applyConfig` (as provided by, e.g., BaseConfigurableClass)
          - `objectName` (as provided by, e.g., QObject)

        :param delegate: (BaseConfigurableClass) The delegate object to be registered.
        :param name: (str) The name to be used as a key for this item in the configuration
                     dictionary. If None given, the object name is used by default.

        .. note:: the registration order will be used when restoring configurations

        .. seealso:: :meth:`unregisterConfigurableItem`, :meth:`registerConfigProperty`, :meth:`createConfig`
        '''
        return self._registerConfigurableItem(delegate, name=name)

    def registerConfigProperty(self, fget, fset, name):
        '''
        Registers a certain property to be included in the config dictionary.

        In this context a "property" is some named value that can be obtained
        via a getter method and can be set via a setter method.

        :param fget: (method or str) method (or name of a method) that gets no
                     arguments and returns the value of a property.
        :param fset: (method or str) method (or name of a method) that gets as
                     an argument the value of a property, and sets it
        :param name: (str) The name to be used as a key for this property in the configuration
                     dictionary

        .. note:: the registration order will be used when restoring configurations

        .. seealso:: :meth:`unregisterConfigurableItem`, :meth:`registerConfigDelegate`, :meth:`createConfig`
        '''
        if isinstance(fget, str) or isinstance(fset, str):
            import weakref
            obj = weakref.proxy(self)
        else:
            obj = None
        p = configurableProperty(name, fget, fset, obj=obj)
        return self._registerConfigurableItem(p, name=name)

    def _registerConfigurableItem(self, item, name=None):
        '''
        Registers the given item as a configurable item which depends of this
        Taurus widget.

        .. note:: This method is not meant to be called directly. Use
                  :meth:`registerConfigProperty`, :meth:`registerConfigDelegate`
                  instead

        Registered items are expected to implement the
        following methods:
          - `createConfig` (as provided by, e.g., BaseConfigurableClass)
          - `applyConfig` (as provided by, e.g., BaseConfigurableClass)
          - `objectName` (as provided by, e.g., QObject)

        :param item: (object) The object that should be registered.
        :param name: (str) The name to be used as a key for this item in the configuration
                     dictionary. If None given, the object name is used by default.

        .. note:: the registration order will be used when restoring configurations

        .. seealso:: :meth:`unregisterConfigurableItem`, :meth:`createConfig`
        '''
        if name is None:
            name = item.objectName()
        name = str(name)
        if name in self.__configurableItemNames:
            # abort if duplicated names
            raise ValueError(
                '_registerConfigurableItem: An object with name "%s" is already registered' % name)
        self.__configurableItemNames.append(name)
        self.__configurableItems[name] = item

    def unregisterConfigurableItem(self, item, raiseOnError=True):
        '''
        unregisters the given item (either a delegate or a property) from the
        configurable items record. It raises an exception if the item is not
        registered

        :param item: (object or str) The object that should be unregistered.
                     Alternatively, the name under which the object was registered
                     can be passed as a python string.
        :param raiseOnError: (bool) If True (default), it raises a KeyError
                             exception if item was not registered. If False, it
                             just logs a debug message

        .. seealso:: :meth:`registerConfigProperty`, :meth:`registerConfigDelegate`
        '''
        if isinstance(item, basestring):
            name = str(item)
        else:
            name = str(item.objectName())
        if name in self.__configurableItemNames and name in self.__configurableItems:
            self.__configurableItemNames.remove(name)
            self.__configurableItems.pop(name)
            return True
        elif raiseOnError:
            raise KeyError('"%s" was not registered.' % name)
        else:
            self.debug('"%s" was not registered. Skipping' % name)
            return False

    def checkConfigVersion(self, configdict, showDialog=False, supportedVersions=None):
        '''
        Check if the version of configdict is supported. By default, the
        BaseConfigurableClass objects have ["__UNVERSIONED__"] as their list of supported
        versions, so unversioned config dicts will be accepted.

        :param configdict: (dict) configuration dictionary to check
        :param showDialog: (bool) whether to show a QtWarning dialog if check
                           failed (false by default)
        :param supportedVersions: (sequence<str>, or None) supported version
                                  numbers, if None given, the versions supported
                                  by this widget will be used (i.e., those
                                  defined in self._supportedConfigVersions)

        :return: (bool) returns True if the configdict is of the right version
        '''
        if supportedVersions is None:
            supportedVersions = self._supportedConfigVersions
        version = configdict.get("ConfigVersion", "__UNVERSIONED__")
        if version not in supportedVersions:
            msg = 'Unsupported Config Version %s. (Supported: %s)' % (
                version, repr(supportedVersions))
            self.warning(msg)
            if showDialog:
                from taurus.external.qt import Qt
                Qt.QMessageBox.warning(
                    self, "Wrong Configuration Version", msg, Qt.QMessageBox.Ok)
            return False
        return True

    def createQConfig(self):
        '''
        returns the current configuration status encoded as a QByteArray. This
        state can therefore be easily stored using QSettings

        :return: (QByteArray) (in the current implementation this is just a
                 pickled configdict encoded as a QByteArray

        .. seealso:: :meth:`restoreQConfig`
        '''
        from taurus.external.qt import Qt
        import cPickle as pickle
        configdict = self.createConfig(allowUnpickable=False)
        return Qt.QByteArray(pickle.dumps(configdict))

    def applyQConfig(self, qstate):
        '''
        restores the configuration from a qstate generated by :meth:`getQState`.

        :param qstate: (QByteArray)

        .. seealso:: :meth:`createQConfig`
        '''
        if qstate.isNull():
            return
        import cPickle as pickle
        configdict = pickle.loads(qstate.data())
        self.applyConfig(configdict)

    def saveConfigFile(self, ofile=None):
        """Stores the current configuration on a file

        :param ofile: (file or string) file or filename to store the configuration

        :return: (str) file name used
        """
        import cPickle as pickle
        if ofile is None:
            from taurus.external.qt import Qt
            ofile = unicode(Qt.QFileDialog.getSaveFileName(
                self, 'Save Configuration', '%s.pck' % self.__class__.__name__, 'Configuration File (*.pck)'))
            if not ofile:
                return
        if not isinstance(ofile, file):
            ofile = open(ofile, 'w')
        configdict = self.createConfig(allowUnpickable=False)
        self.info("Saving current settings in '%s'" % ofile.name)
        pickle.dump(configdict, ofile)
        return ofile.name

    def loadConfigFile(self, ifile=None):
        """Reads a file stored by :meth:`saveConfig` and applies the settings

        :param ifile: (file or string) file or filename from where to read the configuration

        :return: (str) file name used
        """
        import cPickle as pickle
        if ifile is None:
            from taurus.external.qt import Qt
            ifile = unicode(Qt.QFileDialog.getOpenFileName(
                self, 'Load Configuration', '', 'Configuration File (*.pck)'))
            if not ifile:
                return
        if not isinstance(ifile, file):
            ifile = open(ifile, 'r')
        configdict = pickle.load(ifile)
        self.applyConfig(configdict)
        return ifile.name



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
AttributeChooser.py: widget for choosing (a list of) attributes from a tango DB
"""

__all__ = ["TaurusModelSelectorTree", "TaurusModelChooser"]

import sys
from taurus.external.qt import Qt
import taurus.core
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.tree import TaurusDbTreeWidget
from taurus.core.util.containers import CaselessList
from taurusmodellist import TaurusModelList

#@todo: tango-centric!!
from taurus.core.tango.tangodatabase import TangoDevInfo, TangoAttrInfo


class TaurusModelSelectorTree(TaurusWidget):

    addModels = Qt.pyqtSignal('QStringList')

    def __init__(self, parent=None, selectables=None, buttonsPos=None, designMode=None):
        TaurusWidget.__init__(self, parent)
        if selectables is None:
            selectables = [taurus.core.taurusbasetypes.TaurusElementType.Attribute, taurus.core.taurusbasetypes.TaurusElementType.Member,
                           taurus.core.taurusbasetypes.TaurusElementType.Device]
        self._selectables = selectables

        # tree
        self._deviceTree = TaurusDbTreeWidget(
            perspective=taurus.core.taurusbasetypes.TaurusElementType.Device)
        self._deviceTree.getQModel().setSelectables(self._selectables)
        self._deviceTree.setUseParentModel(True)

        # toolbar
        self._toolbar = Qt.QToolBar("TangoSelector toolbar")
        self._toolbar.setIconSize(Qt.QSize(16, 16))
        self._toolbar.setFloatable(False)
        self._addSelectedAction = self._toolbar.addAction(
            Qt.QIcon.fromTheme("list-add"), "Add selected", self.onAddSelected)

        # defines the layout
        self.setButtonsPos(buttonsPos)

        self._deviceTree.recheckTaurusParent()  # NOT WORKING????
        # @todo: This is Workaround because UseSetParentModel is giving trouble again!
        self.modelChanged.connect(self._deviceTree.setModel)

    def setButtonsPos(self, buttonsPos):
        # we must delete the previous layout before we can set a new one
        currlayout = self.layout()
        if currlayout is not None:
            currlayout.deleteLater()
            Qt.QCoreApplication.sendPostedEvents(
                currlayout, Qt.QEvent.DeferredDelete)
        # add to layout
        if buttonsPos is None:
            self.setLayout(Qt.QVBoxLayout())
            self.layout().addWidget(self._deviceTree)
        elif buttonsPos == Qt.Qt.BottomToolBarArea:
            self._toolbar.setOrientation(Qt.Qt.Horizontal)
            self.setLayout(Qt.QVBoxLayout())
            self.layout().addWidget(self._deviceTree)
            self.layout().addWidget(self._toolbar)
        elif buttonsPos == Qt.Qt.TopToolBarArea:
            self._toolbar.setOrientation(Qt.Qt.Horizontal)
            self.setLayout(Qt.QVBoxLayout())
            self.layout().addWidget(self._toolbar)
            self.layout().addWidget(self._deviceTree)
        elif buttonsPos == Qt.Qt.LeftToolBarArea:
            self._toolbar.setOrientation(Qt.Qt.Vertical)
            self.setLayout(Qt.QHBoxLayout())
            self.layout().addWidget(self._toolbar)
            self.layout().addWidget(self._deviceTree)
        elif buttonsPos == Qt.Qt.RightToolBarArea:
            self._toolbar.setOrientation(Qt.Qt.Vertical)
            self.setLayout(Qt.QHBoxLayout())
            self.layout().addWidget(self._deviceTree)
            self.layout().addWidget(self._toolbar)
        else:
            raise ValueError("Invalid buttons position")

    def getSelectedModels(self):
        # todo: this method is tango-centric, but it could be fixed...
        selected = []
        for item in self._deviceTree.selectedItems():
            nfo = item.itemData()
            if isinstance(nfo, TangoDevInfo):
                selected.append(nfo.fullName())
            elif isinstance(nfo, TangoAttrInfo):
                selected.append("%s/%s" %
                                (nfo.device().fullName(), nfo.name()))
            else:
                self.info("Unknown item '%s' in selection" % repr(nfo))
        return selected

    def onAddSelected(self):
        self.addModels.emit(self.getSelectedModels())

    def treeView(self):
        return self._deviceTree.treeView()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        ret['icon'] = "designer:listview.png"
        ret['container'] = False
        ret['group'] = 'Taurus Views'
        return ret


class TaurusModelChooser(TaurusWidget):
    '''A widget that allows the user to select a list of models from a tree representing
    devices and attributes from a Tango server.

    The user selects models and adds them to a list. Then the user should click on the
    update button to notify that the selection is ready.

    signals::
      - "updateModels"  emitted when the user clicks on the update button. It
        passes a list<str> of models that have been selected.
    '''

    updateModels = Qt.pyqtSignal('QStringList')
    UpdateAttrs = Qt.pyqtSignal(['QStringList'], ['QMimeData'])

    def __init__(self, parent=None, selectables=None, host=None, designMode=None, singleModel=False):
        '''Creator of TaurusModelChooser

        :param parent: (QObject) parent for the dialog
        :param selectables: (list<TaurusElementType>) if passed, only elements of the tree whose
                            type is in the list will be selectable.
        :param host: (QObject) Tango host to be explored by the chooser
        :param designMode: (bool) needed for taurusdesigner but ignored here
        :param singleModel: (bool) If True, the selection will be of just one
                            model. Otherwise (default) a list of models can be selected
        '''
        TaurusWidget.__init__(self, parent)
        if host is None:
            host = taurus.Authority().getNormalName()

        self._allowDuplicates = False

        self.setLayout(Qt.QVBoxLayout())

        self.tree = TaurusModelSelectorTree(
            selectables=selectables, buttonsPos=Qt.Qt.BottomToolBarArea)
        self.tree.setModel(host)
        self.list = TaurusModelList()
        self.list.setSelectionMode(Qt.QAbstractItemView.ExtendedSelection)
        applyBT = Qt.QToolButton()
        applyBT.setToolButtonStyle(Qt.Qt.ToolButtonTextBesideIcon)
        applyBT.setText('Apply')
        applyBT.setIcon(Qt.QIcon("status:available.svg"))

        self.setSingleModelMode(singleModel)

        # toolbar
        self._toolbar = self.tree._toolbar
        self._toolbar.addAction(self.list.removeSelectedAction)
        self._toolbar.addAction(self.list.removeAllAction)
        self._toolbar.addAction(self.list.moveUpAction)
        self._toolbar.addAction(self.list.moveDownAction)
        self._toolbar.addSeparator()
        self._toolbar.addWidget(applyBT)
        self.layout().addWidget(self.tree)
        self.layout().addWidget(self.list)

        # self.tree.setUseParentModel(True)  #It does not work!!!!
        # @todo: This is Workaround because UseSetParentModel is giving trouble again!
        self.modelChanged.connect(self.tree.setModel)

        # connections:
        self.tree.addModels.connect(self.addModels)
        applyBT.clicked.connect(self._onUpdateModels)
#        self.connect(self.tree._deviceTree, Qt.SIGNAL("itemDoubleClicked"), self.onTreeDoubleClick)

#    def onTreeDoubleClick(self, item, colum): #@todo: Implement this function properly
#        if item.Role in self.tree._selectables:
#            self.addModels([str(item.text())])

    def getListedModels(self, asMimeData=False):
        '''returns the list of models that have been added

        :param asMimeData: (bool) If False (default), the return value will be a
                           list of models. If True, the return value is a
                           `QMimeData` containing at least `TAURUS_MODEL_LIST_MIME_TYPE`
                           and `text/plain` MIME types. If only one model was selected,
                           the mime data also contains a TAURUS_MODEL_MIME_TYPE.

        :return: (list<str> or QMimeData) the type of return depends on the value of `asMimeData`'''
        models = self.list.getModelList()
        if self.isSingleModelMode():
            models = models[:1]
        if asMimeData:
            md = Qt.QMimeData()
            md.setData(taurus.qt.qtcore.mimetypes.TAURUS_MODEL_LIST_MIME_TYPE, str(
                "\r\n".join(models)))
            md.setText(", ".join(models))
            if len(models) == 1:
                md.setData(
                    taurus.qt.qtcore.mimetypes.TAURUS_MODEL_MIME_TYPE, str(models[0]))
            return md
        return models

    def setListedModels(self, models):
        '''adds the given list of models to the widget list
        '''
        self.list.model().clearAll()
        self.list.addModels(models)

    def resetListedModels(self):
        '''equivalent to setListedModels([])'''
        self.list.model().clearAll()

    def updateList(self, attrList):
        '''for backwards compatibility with AttributeChooser only. Use :meth:`setListedModels` instead'''
        self.info(
            'ModelChooser.updateList() is provided for backwards compatibility only. Use setListedModels() instead')
        self.setListedModels(attrList)

    def addModels(self, models):
        ''' Add given models to the selected models list'''
        if len(models) == 0:
            models = ['']
        if self.isSingleModelMode():
            self.resetListedModels()
        if self._allowDuplicates:
            self.list.addModels(models)
        else:
            listedmodels = CaselessList(self.getListedModels())
            for m in models:
                if m not in listedmodels:
                    listedmodels.append(m)
                    self.list.addModels([m])

    def onRemoveSelected(self):
        '''
        Remove the list-selected models from the list
        '''
        self.list.removeSelected()

    def _onUpdateModels(self):
        models = self.getListedModels()
        self.updateModels.emit(models)
        if taurus.core.taurusbasetypes.TaurusElementType.Attribute in self.tree._selectables:
            # for backwards compatibility with the old AttributeChooser
            self.UpdateAttrs.emit(models)

    def setSingleModelMode(self, single):
        '''sets whether the selection should be limited to just one model
        (single=True) or not (single=False)'''
        if single:
            self.tree.treeView().setSelectionMode(Qt.QAbstractItemView.SingleSelection)
        else:
            self.tree.treeView().setSelectionMode(Qt.QAbstractItemView.ExtendedSelection)
        self._singleModelMode = single

    def isSingleModelMode(self):
        '''returns True if the selection is limited to just one model. Returns False otherwise.

        :return: (bool)'''
        return self._singleModelMode

    def resetSingleModelMode(self):
        '''equivalent to setSingleModelMode(False)'''
        self.setSingleModelMode(self, False)

    @staticmethod
    def modelChooserDlg(parent=None, selectables=None, host=None, asMimeData=False, singleModel=False, windowTitle='Model Chooser'):
        '''Static method that launches a modal dialog containing a TaurusModelChooser

        :param parent: (QObject) parent for the dialog
        :param selectables: (list<TaurusElementType>) if passed, only elements of the tree whose
                            type is in the list will be selectable.
        :param host: (QObject) Tango host to be explored by the chooser
        :param asMimeData: (bool) If False (default),  a list of models will be.
                           returned. If True, a `QMimeData` object will be
                           returned instead. See :meth:`getListedModels` for a
                           detailed description of this QMimeData object.
        :param singleModel: (bool) If True, the selection will be of just one
                            model. Otherwise (default) a list of models can be selected
        :param windowTitle: (str) Title of the dialog (default="Model Chooser")

        :return: (list,bool or QMimeData,bool) Returns a models,ok tuple. models can be
                 either a list of models or a QMimeData object, depending on
                 `asMimeData`. ok is True if the dialog was accepted (by
                 clicking on the "update" button) and False otherwise
        '''
        dlg = Qt.QDialog(parent)
        dlg.setWindowTitle(windowTitle)
        dlg.setWindowIcon(Qt.QIcon("logos:taurus.png"))
        layout = Qt.QVBoxLayout()
        w = TaurusModelChooser(
            parent=parent, selectables=selectables, host=host, singleModel=singleModel)
        layout.addWidget(w)
        dlg.setLayout(layout)
        w.updateModels.connect(dlg.accept)
        dlg.exec_()
        return w.getListedModels(asMimeData=asMimeData), (dlg.result() == dlg.Accepted)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        ret['icon'] = "designer:listview.png"
        ret['container'] = False
        ret['group'] = 'Taurus Views'
        return ret

    singleModelMode = Qt.pyqtProperty(
        "bool", isSingleModelMode, setSingleModelMode, resetSingleModelMode)


def main(args):
    if len(sys.argv) > 1:
        host = sys.argv[1]
    else:
        host = None

    app = Qt.QApplication(args)
    print TaurusModelChooser.modelChooserDlg(host=host)
    sys.exit()

if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a base class for exception hooks"""

__all__ = ["BaseExceptHook"]

__docformat__ = "restructuredtext"

import sys


class BaseExceptHook(object):
    """A callable class that acts as an excepthook that handles an exception.
    This base class simply calls the :obj:`sys.__excepthook__`

    :param hook_to: callable excepthook that will be called at the end of
                    this hook handling [default: None]
    :type hook_to: callable"""

    def __init__(self, hook_to=None):
        self._excepthook = hook_to

    def __call__(self, *exc_info):
        self.report(*exc_info)
        if self._excepthook is not None:
            return self._excepthook(*exc_info)

    def report(self, *exc_info):
        """Report an exception. Overwrite as necessary"""
        return sys.__excepthook__(*exc_info)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
"""

__all__ = ["QObjectRepresentation", "get_qobject_tree", "get_qobject_tree_str",
           "TreeQObjectModel", "TreeQObjectWidget"]

__docformat__ = 'restructuredtext'

import weakref

from taurus.external.qt import Qt

from taurus.core.util.enumeration import Enumeration

QObjectRepresentation = Enumeration('QObjectRepresentation',
                                    ('ClassName', 'ObjectName', 'FullName'))


def _build_qobjects_as_dict(qobject, container):

    container[qobject] = childs = {}
    for child in qobject.children():
        if isinstance(child, Qt.QWidget):
            _build_qobjects_as_dict(child, childs)


def get_qobject_tree_as_dict(qobject=None):

    if qobject is None:
        app = Qt.QApplication.instance()
        qobjects = app.topLevelWidgets()
    else:
        qobjects = [qobject]

    tree = {}
    for qobject in qobjects:
        _build_qobjects_as_dict(qobject, tree)

    return tree


def _build_qobjects_as_list(qobject, container):

    children = qobject.children()
    node = qobject, []
    container.append(node)
    for child in children:
        if isinstance(child, Qt.QWidget):
            _build_qobjects_as_list(child, node[1])


def get_qobject_tree_as_list(qobject=None):

    if qobject is None:
        app = Qt.QApplication.instance()
        qobjects = app.topLevelWidgets()
    else:
        qobjects = [qobject]

    tree = []
    for qobject in qobjects:
        _build_qobjects_as_list(qobject, tree)

    return tree

get_qobject_tree = get_qobject_tree_as_list


def _get_qobject_str(qobject, representation):
    if representation == QObjectRepresentation.ClassName:
        return qobject.__class__.__name__
    elif representation == QObjectRepresentation.ObjectName:
        return str(qobject.objectName())
    elif representation == QObjectRepresentation.FullName:
        return '{0}("{1}")'.format(qobject.__class__.__name__, str(qobject.objectName()))
    return str(qobject)


def _build_qobject_str(node, str_tree, representation=QObjectRepresentation.ClassName):

    qobject, children = node
    str_node = _get_qobject_str(qobject, representation)
    str_children = []
    str_tree.append((str_node, str_children))
    for child in children:
        _build_qobject_str(child, str_children, representation=representation)


def get_qobject_tree_str(qobject=None, representation=QObjectRepresentation.ClassName):

    tree, str_tree = get_qobject_tree(qobject=qobject), []
    for e in tree:
        _build_qobject_str(e, str_tree, representation=representation)
    return str_tree


from taurus.qt.qtgui.tree.qtree import QBaseTreeWidget
from taurus.qt.qtcore.model import TaurusBaseModel, TaurusBaseTreeItem

QR = QObjectRepresentation


class TreeQObjecttItem(TaurusBaseTreeItem):

    def __init__(self, model, data, parent=None):
        TaurusBaseTreeItem.__init__(self, model, data, parent=parent)
        if data is not None:
            self.qobject = weakref.ref(data)
            dat = _get_qobject_str(data, QR.ClassName), \
                _get_qobject_str(data, QR.ObjectName)
            self.setData(0, dat)


class TreeQObjectModel(TaurusBaseModel):

    ColumnNames = "Class", "Object name"
    ColumnRoles = (QR.ClassName,), QR.ObjectName

    def __init__(self, parent=None, data=None):
        TaurusBaseModel.__init__(self, parent=parent, data=data)

#    def createNewRootItem(self):
#        return TreeQObjecttItem(self, self.ColumnNames)

    def role(self, column, depth=0):
        if column == 0:
            return self.ColumnRoles[column][0]
        return self.ColumnRoles[column]

    def roleIcon(self, taurus_role):
        return Qt.QIcon()

    def roleSize(self, taurus_role):
        return Qt.QSize(300, 70)

    def roleToolTip(self, role):
        return "widget information"

    @staticmethod
    def _build_qobject_item(model, parent, node):
        qobject, children = node
        item = TreeQObjecttItem(model, qobject, parent)
        parent.appendChild(item)
        for child in children:
            TreeQObjectModel._build_qobject_item(model, item, child)

    def setupModelData(self, data):
        if data is None:
            return
        rootItem = self._rootItem
        for node in data:
            TreeQObjectModel._build_qobject_item(self, rootItem, node)


class TreeQObjectWidget(QBaseTreeWidget):

    KnownPerspectives = {
        "Default": {
            "label": "Default perspecive",
            "tooltip": "QObject tree view",
            "icon": "",
            "model": [TreeQObjectModel],
        },
    }

    DftPerspective = "Default"

    def __init__(self, parent=None, designMode=False, with_navigation_bar=True,
                 with_filter_widget=True, perspective=None, proxy=None,
                 qobject_root=None):
        QBaseTreeWidget.__init__(self, parent, designMode=designMode,
                                 with_navigation_bar=with_navigation_bar,
                                 with_filter_widget=with_filter_widget,
                                 perspective=perspective, proxy=proxy)
        qmodel = self.getQModel()
        qmodel.setDataSource(get_qobject_tree(qobject=qobject_root))


def build_gui():
    mw = Qt.QMainWindow()
    mw.setObjectName("main window")
    w = Qt.QWidget()
    w.setObjectName("central widget")
    mw.setCentralWidget(w)
    l = Qt.QVBoxLayout()
    w.setLayout(l)
    l1 = Qt.QLabel("H1")
    l1.setObjectName("label 1")
    l.addWidget(l1)
    l2 = Qt.QLabel("H2")
    l2.setObjectName("label 2")
    l.addWidget(l2)
    mw.show()
    return mw


def main():
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication()

    w = build_gui()
    tree = TreeQObjectWidget(qobject_root=w)
    tree.show()
    #import pprint
    # pprint.pprint(get_qobject_tree_str())
    w.dumpObjectTree()
    app.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""event filters library to be used with
:meth:`taurus.qt.qtgui.base.TaurusBaseComponent.setFilters`"""


def IGNORE_ALL(s, t, v):
    '''Will discard all events'''
    return None


def ONLY_CHANGE(s, t, v):
    '''Only change events pass'''
    from taurus.core import TaurusEventType
    if t == TaurusEventType.Change:
        return s, t, v
    else:
        return None


def IGNORE_CHANGE(s, t, v):
    '''Config events are discarded'''
    from taurus.core import TaurusEventType
    if t != TaurusEventType.Change:
        return s, t, v
    else:
        return None


def ONLY_CHANGE_AND_PERIODIC(s, t, v):
    '''Only change events pass'''
    from taurus.core import TaurusEventType
    if t in [TaurusEventType.Change,
             TaurusEventType.Periodic]:
        return s, t, v
    else:
        return None


def IGNORE_CHANGE_AND_PERIODIC(s, t, v):
    '''Config events are discarded'''
    from taurus.core import TaurusEventType
    if t not in [TaurusEventType.Change,
                 TaurusEventType.Periodic]:
        return s, t, v
    else:
        return None


def ONLY_CONFIG(s, t, v):
    '''Only config events pass'''
    from taurus.core import TaurusEventType
    if t == TaurusEventType.Config:
        return s, t, v
    else:
        return None


def IGNORE_CONFIG(s, t, v):
    '''Config events are discarded'''
    from taurus.core import TaurusEventType
    if t != TaurusEventType.Config:
        return s, t, v
    else:
        return None


def IGNORE_FAKE(s, t, v):
    '''Only events with actual value (!=None) pass'''
    if v is not None:
        return s, t, v
    else:
        return None


def ONLY_VALID(s, t, v):
    '''Only events whose quality is VALID pass'''
    from taurus.core import AttrQuality
    if t == AttrQuality.ATTR_VALID:
        return s, t, v
    else:
        return None


class EventValueMap(dict):
    """A filter destined to change the original value into another one according
    to a given map. Example:

        filter = EventValueMap({1:"OPEN", 2:"CHANGING", 3:"CLOSED"})

    this will create a filter that changes the integer value of the event
    into a string. The event type is changed according to the python type in
    the map value.

    For now it only supports simple types: str, int, long, float, bool
    """

    def __call__(self, s, t, v):
        import copy
        from taurus.core import TaurusEventType, DataType

        if not t in (TaurusEventType.Change, TaurusEventType.Periodic):
            return s, t, v
        if v is None:
            return s, t, v

        # make a copy
        v = copy.copy(v)

        v.value = self.get(v.rvalue, v.rvalue)

        v.type = DataType.from_python_type(type(v.rvalue), v.type)
        return s, t, v


class RepeatedEventFilter(object):
    """
    The instances of this class will be callables that can be used as filters
    of repeated-value events.
    If the event type is Change or Periodic, it will only pass when its
    evt_value.value is different from that of the last event received from the
    same source and type. If evt_value.value is not available in the current
    event, the whole evt_value is used for comparison and as a future reference.

    This is useful to avoid processing repetitive events.

    Note that you cannot use this class as a filter: you need to use an
    instance of it.

    Note 2: Use a different instance each time you insert this filter
    into a different widget unless you *really* know what you are doing.

    Example of usage::

        filters = [RepeatedEventFilter(), IGNORE_CONFIG]
        filterEvent(s, t, v, filters)

    """

    def __init__(self):
        self._lastValues = {}

    def __call__(self, s, t, v):
        # restrict this  filter only to change and periodic events.
        if ONLY_CHANGE_AND_PERIODIC(s, t, v) is None:
            return s, t, v
        # block event if we recorded one before with same src, type and v.value
        new_value = getattr(v, 'value', v)
        try:
            if self._lastValues[(s, t)] == new_value:
                return None
        except KeyError:
            pass
        # if it was't blocked, store src, type and v.value for future checks
        self._lastValues[(s, t)] = new_value
        return s, t, v


def filterEvent(evt_src=-1, evt_type=-1, evt_value=-1, filters=()):
    """The event is processed by each and all filters in strict order
    unless one of them returns None (in which case the event is discarded)

    :param evt_src: (object) object that triggered the event
    :param evt_type: (TaurusEventType) type of event
    :param evt_value: (object) event value
    :param filters: (sequence<callable>) a sequence of callables, each returning
                    either None (to discard the event) or the tuple (with
                    possibly transformed values) of
                    (evt_src, evt_type, evt_value)

    :return: (None or tuple) The result of piping the event through the given
             filters.
    """
    evt = evt_src, evt_type, evt_value

    for f in filters:
        evt = f(*evt)
        if evt is None:
            return None
    return evt



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides basic taurus container widget"""

__all__ = ["TaurusWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurusbasecontainer import TaurusBaseContainer


class TaurusWidget(Qt.QWidget, TaurusBaseContainer):
    """This is a Qt.QWidget that additionally accepts a model property.
    This type of taurus container classes are specially useful if you define
    a parent taurus model to them and set all contained taurus widgets to use parent
    model. Example::

        from taurus.qt.qtgui.container import *
        from taurus.qt.qtgui.display import *

        widget = TaurusWidget()
        layout = Qt.QVBoxLayout()
        widget.setLayout(layout)
        widget.model = 'sys/database/2'
        stateWidget = TaurusLabel()
        layout.addWidget(stateWidget)
        stateWidget.useParentModel = True
        stateWidget.model = '/state'"""

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseContainer, name, designMode=designMode)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseContainer.getQtDesignerPluginInfo()
        if cls is TaurusWidget:
            ret['module'] = 'taurus.qt.qtgui.container'
            ret['group'] = 'Taurus Containers'
            ret['icon'] = "designer:frame.png"
            ret['container'] = True
        return ret
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSlot()
    def applyPendingChanges(self):
        self.applyPendingOperations()

    @Qt.pyqtSlot()
    def resetPendingChanges(self):
        self.resetPendingOperations()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseContainer.getModel,
                            TaurusBaseContainer.setModel,
                            TaurusBaseContainer.resetModel)

    useParentModel = Qt.pyqtProperty("bool", TaurusBaseContainer.getUseParentModel,
                                     TaurusBaseContainer.setUseParentModel,
                                     TaurusBaseContainer.resetUseParentModel)

    showQuality = Qt.pyqtProperty("bool", TaurusBaseContainer.getShowQuality,
                                  TaurusBaseContainer.setShowQuality,
                                  TaurusBaseContainer.resetShowQuality)

    modifiableByUser = Qt.pyqtProperty("bool", TaurusBaseContainer.isModifiableByUser,
                                       TaurusBaseContainer.setModifiableByUser,
                                       TaurusBaseContainer.resetModifiableByUser)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import functools


class memoized(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)
